#ifndef __NAME_REGISTRY_H__
#define __NAME_REGISTRY_H__

/* NameRegistry error codes */
enum {
    nrLockedErr				= -2536,
    nrNotEnoughMemoryErr		= -2537,
    nrInvalidNodeErr			= -2538,
    nrNotFoundErr			= -2539,
    nrNotCreatedErr			= -2540,
    nrNameErr				= -2541,
    nrNotSlotDeviceErr			= -2542,
    nrDataTruncatedErr			= -2543,
    nrPowerErr				= -2544,
    nrPowerSwitchAbortErr		= -2545,
    nrTypeMismatchErr			= -2546,
    nrNotModifiedErr			= -2547,
    nrOverrunErr			= -2548,
    nrResultCodeBase			= -2549,
    nrPathNotFound			= -2550,    /* a path component lookup failed */
    nrPathBufferTooSmall		= -2551,    /* buffer for path is too small */
    nrInvalidEntryIterationOp		= -2552,    /* invalid entry iteration operation */
    nrPropertyAlreadyExists		= -2553,    /* property already exists */
    nrIterationDone			= -2554,    /* iteration operation is done */
    nrExitedIteratorScope		= -2555,    /* outer scope of iterator was exited */
    nrTransactionAborted		= -2556,        /* transaction was aborted */

    gestaltUndefSelectorErr		= -5551 /*undefined selector was passed to Gestalt*/
};

enum {
    kNVRAMProperty        		= 0x00000020L,            // matches NR
};


struct RegEntryID
{
    void * opaque[4];
};

typedef struct RegEntryID		RegEntryID;
typedef RegEntryID *                    RegEntryIDPtr;
typedef void *                          RegPropertyValue;
typedef UInt32                          RegPropertyValueSize;

/* Max length of Entry Name */
enum {
  kRegCStrMaxEntryNameLength    = 47
};

/* Entry Names are single byte ASCII */
typedef char                            RegCStrEntryName;
typedef char *                          RegCStrEntryNamePtr;
/* length of RegCStrEntryNameBuf =  kRegCStrMaxEntryNameLength+1*/
typedef char                            RegCStrEntryNameBuf[48];
typedef char                            RegCStrPathName;
typedef UInt32                          RegPathNameSize;
enum {
  kRegPathNameSeparator         = ':',  /* 0x3A */
  kRegEntryNameTerminator       = 0x00, /* '\0' */
  kRegPathNameTerminator        = 0x00  /* '\0' */
};

enum {
  kRegMaximumPropertyNameLength = 31,   /* Max length of Property Name */
  kRegPropertyNameTerminator    = 0x00  /* '\0' */
};

typedef char                            RegPropertyNameBuf[32];
typedef char                            RegPropertyName;
typedef char *                          RegPropertyNamePtr;
enum {
  kRegMaxPropertyNameLength     = kRegMaximumPropertyNameLength
};

typedef struct _RegEntryIter * RegEntryIter;
typedef struct _RegPropertyIter * RegPropertyIter;

typedef UInt32                          RegIterationOp;
typedef RegIterationOp                  RegEntryIterationOp;
enum {
                                        /* Absolute locations*/
  kRegIterRoot                  = 0x00000002, /* "Upward" Relationships */
  kRegIterParents               = 0x00000003, /* include all  parent(s) of entry */
                                        /* "Downward" Relationships*/
  kRegIterChildren              = 0x00000004, /* include all children */
  kRegIterSubTrees              = 0x00000005, /* include all sub trees of entry */
  kRegIterDescendants           = 0x00000005, /* include all descendants of entry */
                                        /* "Horizontal" Relationships */
  kRegIterSibling               = 0x00000006, /* include all siblings */
                                        /* Keep doing the same thing*/
  kRegIterContinue              = 0x00000001
};

typedef UInt32                          RegModifiers;
typedef RegModifiers                    RegEntryModifiers;
typedef RegModifiers                    RegPropertyModifiers;
enum {
  kRegNoModifiers               = 0x00000000, /* no entry modifiers in place */
  kRegUniversalModifierMask     = 0x0000FFFF, /* mods to all entries */
  kRegNameSpaceModifierMask     = 0x00FF0000, /* mods to all entries within namespace */
  kRegModifierMask              = (long)0xFF000000 /* mods to just this entry */
};

/* Universal Property Modifiers */
enum {
  kRegPropertyValueIsSavedToNVRAM = 0x00000020, /* property is non-volatile (saved in NVRAM) */
  kRegPropertyValueIsSavedToDisk = 0x00000040 /* property is non-volatile (saved on disk) */
};


#endif /* __NRLIB_H__ */
