#include "ppc_asm.h"

//#define TRACE_LIB_CALLS
#undef TRACE_LIB_CALLS
	
#define ENTRY(name) \
	.text ; \
	.align 2 ; \
	.globl name; \
	name:


/*
 * So, this is a bit tricky. We are coming from Linux SysV ABI and calling
 * into MacOS PEF ABI. We create a dummy parameter area of 64 bytes that
 * just for fun, and a small linkage area
 *
 * We intentionally break the linkage between the linux back chain and
 * the MacOS one here.
 *
 * Parameters are in r3...r8, function pointer in r9
 */
#define PEF_FRAME_PARAM_AREA_SIZE	64
#define PEF_FRAME_LINKAGE_AREA_SIZE	24
#define PEF_FRAME_SIZE			(PEF_FRAME_PARAM_AREA_SIZE + \
					 PEF_FRAME_LINKAGE_AREA_SIZE)

#define PEF_LINKAGE_SP			0
#define PEF_LINKAGE_CR			4	
#define PEF_LINKAGE_LR			8	
#define PEF_LINKAGE_TOC			20	

ENTRY(CallTVector)
	/* Save LR in the linux frame */
	mflr	r0
	stw	r0,4(r1)

	/* Move down stack pointer, write 0 to back link */
	subi	r1,r1,PEF_FRAME_SIZE
	li	r0,0
	stw	r0,0(r1)

	/* Save linux r2 just in case it contained something
	 * meaningful, maybe useless though, we know we aren't
	 * TLS compatible ;)
	 */
	stw	r2,PEF_LINKAGE_TOC(r1)
	
	/* Get the function pointer and TOC out of the TVector,
	 * and do the call
	 */
	lwz	r0,0(r9)
	lwz	r2,4(r9)
	mtlr	r0
	blrl

	/* Pop things back */
	lwz	r2,PEF_LINKAGE_TOC(r1)
	addi	r1,r1,PEF_FRAME_SIZE
	lwz	r0,4(r1)
	mtlr	r0
	blr
	

/*
 * So here, now, we are coming from PEF and calling into Linux code,
 * like above, we only care about register based parameters at this
 * point. We create an empty stack frame after having saved the link
 * register of the caller
 */
ENTRY(LibraryTrampoline)
	mflr	r0
	stw	r0,PEF_LINKAGE_LR(r1)
	mfcr	r0
	stw	r0,PEF_LINKAGE_CR(r1)

	/* Create linux stackframe with arbitrary size for now */
	li	r0, 0
	stwu	r0,-128(r1)

	/* Recover function pointer in r9, store parameter
	 * block in r10
	 */
	mr	r10,r2
	lwz	r9,12(r10)

	/* Restore linux initial r2 value */
	lis	r11,linux_r2@ha
	lwz	r2,linux_r2@l(r11)

#ifdef TRACE_LIB_CALLS
	stw	r3,8(r1)
	stw	r4,12(r1)
	stw	r5,16(r1)
	stw	r6,20(r1)
	stw	r7,24(r1)
	stw	r8,28(r1)
	stw	r9,32(r1)
	mr	r3,r10
	mflr	r4
	addi	r5,r1,8
	bl	libcall_trace
	lwz	r3,8(r1)
	lwz	r4,12(r1)
	lwz	r5,16(r1)
	lwz	r6,20(r1)
	lwz	r7,24(r1)
	lwz	r8,28(r1)
	lwz	r9,32(r1)
#endif /* TRACE_LIB_CALLS */

	/* Do the jump */
	mtlr	r9
	blrl

	/* Pop back */
	addi	r1,r1,128
	lwz	r0,PEF_LINKAGE_CR(r1)
	mtcr	r0
	lwz	r0,PEF_LINKAGE_LR(r1)
	mtlr	r0
	blr

ENTRY(InitLinuxR2)
	lis	r3,linux_r2@ha
	stw	r2,linux_r2@l(r3)
	blr
	
	.data
	.align 2
linux_r2:
	.long 0

/*
 * Here, we flush the dcache & invalidate the icache for a given
 * range (making data executable)
 */

/* Possibly incorrect assumption but will work well enough for now */
#define L1_CACHE_LINE_SIZE	32
#define LG_L1_CACHE_LINE_SIZE	5
#define SPRN_PVR		0x11F

ENTRY(MakeDataExecutable)
	mfspr	r5,SPRN_PVR		/* kernel is kind enough to emulate that
					 * insn for us */
	rlwinm	r5,r5,16,16,31
	cmpi	0,r5,1
	beqlr				/* for 601, do nothing */
	add	r4,r3,r4
	subi	r4,r4,1
	li	r5,L1_CACHE_LINE_SIZE-1
	andc	r3,r3,r5
	subf	r4,r3,r4
	add	r4,r4,r5
	srwi.	r4,r4,LG_L1_CACHE_LINE_SIZE
	beqlr
	mtctr	r4
	mr	r6,r3
1:	dcbst	0,r3
	addi	r3,r3,L1_CACHE_LINE_SIZE
	bdnz	1b
	sync				/* wait for dcbst's to get to ram */
	mtctr	r4
2:	icbi	0,r6
	addi	r6,r6,L1_CACHE_LINE_SIZE
	bdnz	2b
	sync				/* additional sync needed on g4 */
	isync
	blr
