/*
 * Copyright (c) 1998-2000 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * Copyright (c) 1999-2003 Apple Computer, Inc.  All Rights Reserved.
 * 
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 * 
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>

#include "ndrver.h"

static inline void eieio(void)
{
	asm volatile ("eieio" : : : "memory");
}

static inline int compare_and_swap(UInt32 old, UInt32 new, UInt32 *vptr)       
{
	UInt32 prev;

	asm volatile ("\n\
	1:	lwarx	%0,0,%2;\n\
		cmpw	0,%0,%3;\n\
		bne	2f;\n\
		stwcx.	%4,0,%2;\n\
		bne-	1b;\n\
		sync;\n\
	2:	"     : "=&r" (prev), "=m" (*vptr)
		      : "r" (vptr), "r" (old), "r" (new), "m" (*vptr)
		      : "cc", "memory");
	return (prev == old);
}

static void udelay(long usec)
{
	struct timeval tv1, tv2;
	struct timezone tz;
	long long t1, t2;
	
//	LOG_DBG("DelayForHardware: %d usec\n", time);
	if (usec > 1000000)
		usec = 1000000;
	gettimeofday(&tv1, &tz);
	t1 = tv1.tv_sec * 1000000 + tv1.tv_usec;

	for (;;) {
		gettimeofday(&tv2, &tz);
		t2 = tv2.tv_sec * 1000000 + tv2.tv_usec;
		if ((t2 - t1) > usec)
			break;
	}
}

#define EXP(s)	s


OSStatus EXP(VSLSetDisplayConfiguration)(RegEntryID * entryID,
					char *	propertyName,
					void *	configData,
					long	configDataSize)
{
	return -1;
}

enum {
    kProcessorCacheModeDefault		= 0,
    kProcessorCacheModeInhibited 	= 1,
    kProcessorCacheModeWriteThrough 	= 2,
    kProcessorCacheModeCopyBack		= 3
};

OSStatus EXP(SetProcessorCacheMode)( UInt32 space, void * addr,
				    UInt32 len, UInt32 mode)
{
    eieio();
    return noErr;
}

Boolean EXP(CompareAndSwap)(
  UInt32    oldVvalue,
  UInt32    newValue,
  UInt32 *  OldValueAdr)
{
    return compare_and_swap(oldVvalue, newValue, OldValueAdr);
}

void EXP(SynchronizeIO)(void)
{
    eieio();
}

void EXP(BlockCopy)(
  const void *  srcPtr,
  void *        destPtr,
  Size          byteCount)
{
	/* This may be called on non-cacheable memory, beware of what
	 * implementation is used
	 */
	//	LOG_DBG(" BlockCopy(src: %p, dst: %p, size: %ld\n", srcPtr, destPtr, byteCount);
	memcpy(destPtr, srcPtr, byteCount);
}

void EXP(BlockMove)(
  const void *  srcPtr,
  void *        destPtr,
  Size          byteCount)
{
	/* This may be called on non-cacheable memory, beware of what
	 * implementation is used
	 */
	//	LOG_DBG(" BlockMove(src: %p, dst: %p, size: %ld\n", srcPtr, destPtr, byteCount);
	memmove(destPtr, srcPtr, byteCount);
}

void EXP(BlockMoveData)(
  const void *  srcPtr,
  void *        destPtr,
  Size          byteCount)
{
	/* This may be called on non-cacheable memory, beware of what
	 * implementation is used
	 */
	memmove(destPtr, srcPtr, byteCount);
}

void EXP(BlockMoveDataUncached)(
  const void *  srcPtr,
  void *        destPtr,
  Size          byteCount)
{
	/* This may be called on non-cacheable memory, beware of what
	 * implementation is used
	 */
	memmove(destPtr, srcPtr, byteCount);
}

void EXP(BlockMoveUncached)(
  const void *  srcPtr,
  void *        destPtr,
  Size          byteCount)
{
	/* This may be called on non-cacheable memory, beware of what
	 * implementation is used
	 */
	memmove(destPtr, srcPtr, byteCount);
}

void EXP(BlockZero)(
  void *  srcPtr,
  Size          byteCount)
{
	/* This may be called on non-cacheable memory, beware of what
	 * implementation is used
	 */
	memset(srcPtr, 0, byteCount);
}

void EXP(BlockZeroUncached)(
  void *  srcPtr,
  Size          byteCount)
{
	/* This may be called on non-cacheable memory, beware of what
	 * implementation is used
	 */
	memset(srcPtr, 0, byteCount);
}

char * EXP(CStrCopy)( char * dst, const char * src)
{
	return strcpy(dst,src);
}

SInt16 EXP(CStrCmp)(const char *s1, const char *s2)
{
	return strcmp(s1, s2);
}

UInt32 EXP(CStrLen)(const char * src)
{
	return strlen(src);
}

char * EXP(CStrCat)(char *dst, const char *src)
{
	return strcat(dst, src);
}

char * EXP(CStrNCopy)(char *dst, const char *src, UInt32 max)
{
	return strncpy(dst, src, max);
}

SInt16 EXP(CStrNCmp)(const char *s1, const char *s2, UInt32 max)
{
    return strncmp(s1, s2, max);
}

char * EXP(CStrNCat)(char *dst, const char *src, UInt32 max)
{
    return strncat(dst, src, max);
}

char * EXP(PStrCopy)(char *to, const char *from)
{
	UInt32	len;
	char   *copy;

	copy = to;
	len = *(from++);
	*(copy++) = len;
	memcpy(copy, from, len);

	return to;
}

void EXP(PStrToCStr)(char *to, const char *from)
{
	UInt32	len;

	len = *(from++);
	memcpy(to, from, len);
	*(to + len) = 0;
}

void EXP(CStrToPStr)(char *to, const char *from)
{
	UInt32	len;

	len = strlen(from);
	*to = len;
	memcpy(to + 1, from, len);
}

LogicalAddress EXP(PoolAllocateResident)(ByteCount byteSize, Boolean clear)
{
	UInt32 * mem;

	mem = (UInt32 *)malloc((size_t)byteSize + sizeof(UInt32) );
	if (mem == NULL)
		return NULL;

	mem[0] = 'mllc';
	mem++;

	if (clear)
		memset(mem, 0, byteSize);

	LOG_DBG(" PoolAllocateResident(%ld, %d) -> %p\n", byteSize, clear, mem);
	return (LogicalAddress)mem;
}

OSStatus EXP(PoolDeallocate)( LogicalAddress address )
{
	UInt32 * mem = (UInt32 *) address;

	if (!mem)
		return -108;

	LOG_DBG(" PoolDeallocate(%p)\n", mem);

	mem--;
	if (mem[0] != 'mllc') {
		IOLog("PoolDeallocate invalid address %08lx\n", mem + 1);
		return -1;
	}
	free( (void *) mem );

	return noErr;
}

UInt32	EXP(CurrentExecutionLevel)(void)
{
	/* TODO: Implement interrupts as signals */
	return 0;
#if 0
    if (ml_at_interrupt_context())
	return (6);		// == kTaskLevel, HWInt == 6
    else
	return (0);		// == kTaskLevel, HWInt == 6
#endif
}

// don't expect any callers of this
OSErr EXP(IOCommandIsComplete)( IOCommandID commandID, OSErr result)
{
	IOLog("IOCommandIsComplete\n");
	return result;		// !!??!!
}


/*
 * Time management: TODO: Use mftb in userland and retreive calibration
 * infos from the kernel instead
 */

/*AbsoluteTime EXP(UpTime)( void )*/
void UpTime(AbsoluteTime *result)
{
	struct timeval tv;
	struct timezone tz;

	gettimeofday(&tv, &tz);

	*result = ((UInt64)tv.tv_sec * 1000000) + (UInt64)tv.tv_usec;
}

/*AbsoluteTime EXP(AddAbsoluteToAbsolute)(AbsoluteTime left, AbsoluteTime right)*/
void AddAbsoluteToAbsolute(AbsoluteTime *result , AbsoluteTime left, AbsoluteTime right)
{
	*result = left + right;
}


/*AbsoluteTime EXP(SubAbsoluteFromAbsolute)(AbsoluteTime left, AbsoluteTime right)*/
void SubAbsoluteFromAbsolute(AbsoluteTime *result, AbsoluteTime left, AbsoluteTime right)
{
	LOG_DBG ("  SubAbsoluteFromAbsolute: result @%p, left: %d, right: %d\n",
		 result, (int)left, (int)right);
    // !! ATI bug fix here:
    // They expect the 64-bit result to be signed. The spec says < 0 => 0
    // To workaround, make sure this routine takes 10 us to execute.
    udelay(10);

    // XXX Check the order of arguments
    *result = (left < right) ? 0 : left - right;
}


/*AbsoluteTime*/void    EXP(DurationToAbsolute)( /*Duration theDuration */
					AbsoluteTime *result, int theDuration)
{
	unsigned long absolute;

	absolute = (theDuration > 0) ? (theDuration * 1000) : -theDuration;
	//	LOG_DBG("  DurationToAbsolute: (%08x,%08x) -> %08x\n", result, theDuration, (int)absolute);
	//	return (AbsoluteTime)absolute;
	*result = absolute;
}

/*AbsoluteTime EXP(AddDurationToAbsolute)( Duration duration, AbsoluteTime absolute ) */
void AddDurationToAbsolute( AbsoluteTime *result, Duration duration, AbsoluteTime absolute )
{
	if (duration > 0)
		duration *= 1000;
	else
		duration = -duration;
	*result = absolute + (UInt64)duration;
}

//#define UnsignedWideToUInt64(x)		(*(UInt64 *)(x))
//#define UInt64ToUnsignedWide(x)		(*(UnsignedWide *)(x))

/*AbsoluteTime    EXP(NanosecondsToAbsolute) ( UnsignedWide theNanoseconds)*/
void NanosecondsToAbsolute ( AbsoluteTime *result, UnsignedWide theNanoseconds)
{
	*result = *((UInt64 *)&theNanoseconds);
}

/*UnsignedWide    EXP(AbsoluteToNanoseconds)( AbsoluteTime absolute )*/
void AbsoluteToNanoseconds( UnsignedWide *result, AbsoluteTime absolute )
{
	UnsignedWide r;

	*((UInt64 *)&r) = absolute;
	*result = r;
}

Duration    EXP(AbsoluteDeltaToDuration)( AbsoluteTime left, AbsoluteTime right )
{
#if 0
    Duration		dur;
    AbsoluteTime	result;
    UInt64		nano;

    if (CMP_ABSOLUTETIME(&left, &right) < 0)
	return (0);

    result = left;
    SUB_ABSOLUTETIME( &result, &right);
    absolutetime_to_nanoseconds( result, &nano);

    if (nano >= ((1ULL << 31) * 1000ULL))
    {
	// +ve milliseconds
	if (nano >= ((1ULL << 31) * 1000ULL * 1000ULL))
	    dur = 0x7fffffff;
	else
	    dur = nano / 1000000ULL;
    }
    else
    {
	// -ve microseconds
	dur = -(nano / 1000ULL);
    }

    return (dur);
#endif
    return 0;
}

Duration    EXP(AbsoluteToDuration)( AbsoluteTime result )
{
#if 0
    Duration		dur;
    UInt64		nano;

    absolutetime_to_nanoseconds( result, &nano);

    if (nano >= ((1ULL << 31) * 1000ULL))
    {
	// +ve milliseconds
	if (nano >= ((1ULL << 31) * 1000ULL * 1000ULL))
	    dur = 0x7fffffff;
	else
	    dur = nano / 1000000ULL;
    }
    else
    {
	// -ve microseconds
	dur = -(nano / 1000ULL);
    }

    return (dur);
#endif
    return 0;
}

OSStatus    EXP(DelayForHardware)(AbsoluteTime time)
{
#if 1
	if (time > 10)
		LOG_DBG("  DelayForHardware: %d\n", (int)time);
#endif
	udelay((long)time);
	return noErr;
}

OSStatus    EXP(DelayFor)( Duration theDuration )
{
#if 0
#if 1

    // In Marconi, DelayFor uses the old toolbox Delay routine
    // which is based on the 60 Hz timer. Durations are not
    // rounded up when converting to ticks. Yes, really.
    // Some ATI drivers call DelayFor(1) 50000 times starting up.
    // There is some 64-bit math there so we'd better reproduce
    // the overhead of that calculation.

#define DELAY_FOR_TICK_NANO		16666666
#define DELAY_FOR_TICK_MILLI		17
#define NANO32_MILLI			4295

    UnsignedWide	nano;
    AbsoluteTime	abs;
    unsigned int	ms;

    abs = EXP(DurationToAbsolute)( theDuration);
    nano = EXP(AbsoluteToNanoseconds)( abs);

    ms = (nano.lo / DELAY_FOR_TICK_NANO) * DELAY_FOR_TICK_MILLI;
    ms += nano.hi * NANO32_MILLI;
    if (ms)
	IOSleep( ms);

#else
    // Accurate, but incompatible, version

#define SLEEP_THRESHOLD		5000

    if (theDuration < 0)
    {
	// us duration
	theDuration -= theDuration;
	if (theDuration > SLEEP_THRESHOLD)
	    IOSleep( (theDuration + 999) / 1000);
	else
	    IODelay( theDuration);
    }
    else
    {
	// ms duration
	if (theDuration > (SLEEP_THRESHOLD / 1000))
	    IOSleep( theDuration );                     	// ms
	else
	    IODelay( theDuration * 1000);			// us
    }
#endif
#endif
    return noErr;
}

void EXP(SysDebug)(void)
{
	IOLog("**** SysDebug ****\n");
}

void EXP(SysDebugStr)( const char * from )
{
	IOLog("**** SysDebugStr ****\n%s", from);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static IOReturn
ApplePMUSendMiscCommand( UInt32 command,
			    IOByteCount sendLength, UInt8 * sendBuffer,
			    IOByteCount * readLength, UInt8 * readBuffer )
{
#if 0
    struct SendMiscCommandParameterBlock
    {
	int command;
	IOByteCount sLength;
	UInt8 *sBuffer;
	IOByteCount *rLength;
	UInt8 *rBuffer;
    };
    IOReturn ret = kIOReturnError;
    static IOService * pmu;

    // See if ApplePMU exists
    if (!pmu)
    {
	OSIterator * iter;
	iter = IOService::getMatchingServices(IOService::serviceMatching("ApplePMU"));
	if (iter)
	{
	    pmu = (IOService *) iter->getNextObject();
	    iter->release();
	}
    }

    SendMiscCommandParameterBlock params = { command, sendLength, sendBuffer,
					    readLength, readBuffer };
    if (pmu)
	ret = pmu->callPlatformFunction( "sendMiscCommand", true,
					    (void*)&params, NULL, NULL, NULL );
    return (ret);
#endif
    return -1;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#define	readExtSwitches	0xDC

extern IOOptionBits gIOFBLastClamshellState;
extern bool	    gIOFBDesktopModeAllowed;

OSStatus EXP(VSLGestalt)( VSLGestaltType selector, UInt32 * response )
{
#if 0
    IOReturn ret;

    if (!response)
	return (paramErr);

    *response = 0;

    switch (selector)
    {
	case kVSLClamshellStateGestaltType:
	    {
		UInt8 bootEnvIntData[32];
		IOByteCount iLen = sizeof(UInt8);

		ret = ApplePMUSendMiscCommand(readExtSwitches, 0, NULL, &iLen, &bootEnvIntData[0]);
		if (kIOReturnSuccess == ret)
		{
		    gIOFBLastClamshellState = bootEnvIntData[0] & 1;
		    if (gIOFBDesktopModeAllowed)
			*response = bootEnvIntData[0];
		}
		break;
	    }
	default:
	    ret = gestaltUndefSelectorErr;
	    break;
    }

    return (ret);
#endif
    return -1;
}

OSErr
EXP(VSLNewInterruptService)(
  RegEntryID *            serviceDevice,
  InterruptServiceType    serviceType,
  InterruptServiceIDPtr   serviceID)
{
	return -1;
}

OSErr
EXP(VSLDisposeInterruptService)(InterruptServiceIDType serviceID)
{
	return -1;
}

OSErr
EXP(VSLDoInterruptService)(InterruptServiceIDType serviceID)
{
	return -1;
}

Boolean
EXP(VSLPrepareCursorForHardwareCursor)(
  void *                        cursorRef,
  HardwareCursorDescriptorPtr   hardwareDescriptor,
  HardwareCursorInfoPtr         hwCursorInfo)
{
	return -1;
}

OSStatus _eCallOSTrapUniversalProc( UInt32 theProc,
				    UInt32 procInfo, UInt32 trap, UInt8 * pb )
{
#if 0
    OSStatus    err = -40;
#ifdef __ppc__
    struct PMgrOpParamBlock
    {
	SInt16	pmCommand;
	SInt16	pmLength;
	UInt8 *	pmSBuffer;
	UInt8 *	pmRBuffer;
	UInt8	pmData[4];
    };
    UInt8	bootEnvIntData[32];

    if ((procInfo == 0x133822)
	    && (trap == 0xa085))
    {
	PMgrOpParamBlock * pmOp = (PMgrOpParamBlock *) pb;

	if ((readExtSwitches == pmOp->pmCommand) && pmOp->pmRBuffer)
	{
#if 1
	    IOReturn ret;
	    IOByteCount iLen = sizeof(UInt8);
	    ret = ApplePMUSendMiscCommand(readExtSwitches, 0, NULL, &iLen, &bootEnvIntData[0]);
	    if (kIOReturnSuccess == ret)
		*pmOp->pmRBuffer = (bootEnvIntData[0] & 1);
	    else
#endif

	    {
		OSNumber * num = OSDynamicCast(OSNumber,
						IOService::getPlatform()->getProperty("AppleExtSwitchBootState"));
		if (num)
		    *pmOp->pmRBuffer = (num->unsigned32BitValue() & 1);
		else
		    *pmOp->pmRBuffer = 0;
	    }
	    gIOFBLastClamshellState = *pmOp->pmRBuffer;
	    err = noErr;
	}
    }
    else if ((procInfo == 0x133822)
		&& (trap == 0xa092))
    {
	UInt8 addr, reg, data;

	addr = pb[ 2 ];
	reg = pb[ 3 ];
	pb = *( (UInt8 **) ((UInt32) pb + 8));
	data = pb[ 1 ];
	(*PE_write_IIC)( addr, reg, data );
	err = noErr;
    }
#endif
    return err;
#endif
    return noErr;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

const UInt32 * _eGetKeys( void )
{
	static const UInt32 zeros[] = { 0, 0, 0, 0 };

	return zeros;
}

UInt32 _eGetIndADB( void * adbInfo, UInt32 index)
{
	memset(adbInfo, 0, 10);
	return 0;		// orig address
}

char * _eLMGetPowerMgrVars( void )
{
	static char * powerMgrVars = NULL;

	if (powerMgrVars == NULL) {
		powerMgrVars = (char *)malloc( 0x3c0);
		if (powerMgrVars)
			memset(powerMgrVars, 0, 0x3c0);
	}
	return powerMgrVars;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

OSStatus _eNoErr( void )
{
	return noErr;
}

OSStatus _eFail( void )
{
	return -40;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

// fix this!

#define 	heathrowID		((volatile UInt32 *)0xf3000034)
#define 	heathrowTermEna		(1 << 3)
#define 	heathrowTermDir		(1 << 0)

#define 	heathrowFeatureControl	((volatile UInt32 *)0xf3000038)
#define 	heathrowMBRES		(1 << 24)

#define 	heathrowBrightnessControl ((volatile UInt8 *)0xf3000032)
#define		defaultBrightness	144
#define 	heathrowContrastControl ((volatile UInt8 *)0xf3000033)
#define		defaultContrast		183

#define 	gossamerSystemReg1	((volatile UInt16 *)0xff000004)
#define		gossamerAllInOne	(1 << 4)

void EXP(ATISetMBRES)( UInt32 state )
{
#if 0
    UInt32	value;

    value = *heathrowFeatureControl;

    if (state == 0)
	value &= ~heathrowMBRES;
    else if (state == 1)
	value |= heathrowMBRES;

    *heathrowFeatureControl = value;
    eieio();
#endif
}

void EXP(ATISetMonitorTermination)( Boolean enable )
{
#if 0
    UInt32	value;

    value = *heathrowID;

    value |= heathrowTermEna;
    if (enable)
	value |= heathrowTermDir;
    else
	value &= ~heathrowTermDir;

    *heathrowID = value;
    eieio();
#endif
}

Boolean EXP(ATIIsAllInOne)( void )
{
#if 0
    Boolean	rtn;
    static bool	didBrightness;

    rtn = (0 == ((*gossamerSystemReg1) & gossamerAllInOne));
    if (rtn && !didBrightness)
    {
	*heathrowBrightnessControl = defaultBrightness;
	eieio();
	*heathrowContrastControl = defaultContrast;
	eieio();
	didBrightness = true;
    }
    return (rtn);
#endif
    return 0;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#if 0
static void IONDRVInterruptAction( OSObject * target, void * refCon,
				    IOService * provider, int index )
{
    IONDRVInterruptSet *	set;
    IONDRVInterruptSource *	source;
    SInt32			result;

    set = (IONDRVInterruptSet *) target;
    index++;

    do
    {
	assert( (UInt32) index <= set->count);
	if ((UInt32) index > set->count)
		break;

	source = set->sources + index;
	result = CallTVector( set, (void *) index, source->refCon, 0, 0, 0,
				    source->handler );

	switch (result)
	{
	    case kIONDRVIsrIsNotComplete:
		index++;
	    case kIONDRVIsrIsComplete:
		break;

	    case kIONDRVMemberNumberParent:
		assert( false );
		break;

	    default:
		index = result;
		set = set->child;
		break;
	}
    }
    while (result != kIONDRVIsrIsComplete);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static SInt32 IONDRVStdInterruptHandler( IONDRVInterruptSetMember setMember,
	void *refCon, UInt32 theIntCount )
{
    //    assert( false );

    return (kIONDRVIsrIsComplete);
}

static bool IONDRVStdInterruptDisabler( IONDRVInterruptSetMember setMember,
					void *refCon )
{
    IONDRVInterruptSet *	set;
    IONDRVInterruptSource *	source;
    bool			was;

    set = (IONDRVInterruptSet *) setMember.setID;
    assert( OSDynamicCast( IONDRVInterruptSet, set ));
    assert( setMember.member <= set->count );
    source = set->sources + setMember.member;

    was = source->enabled;
    source->enabled = false;

    assert( set->provider );
    set->provider->disableInterrupt( setMember.member - 1 );

    return (was);
}

static void IONDRVStdInterruptEnabler( IONDRVInterruptSetMember setMember,
					void *refCon )
{
    IONDRVInterruptSet *	set;
    IONDRVInterruptSource *	source;

    set = (IONDRVInterruptSet *) setMember.setID;
    assert( OSDynamicCast( IONDRVInterruptSet, set ));
    assert( setMember.member <= set->count );
    source = set->sources + setMember.member;

    source->enabled = true;

    assert( set->provider );

    if (!source->registered)
    {
	source->registered = true;
	set->provider->registerInterrupt( setMember.member - 1, set,
						    &IONDRVInterruptAction, (void *) 0x53 );
    }

    set->provider->enableInterrupt( setMember.member - 1 );
}

/// FIXME: Use proper branch islands
static IOTVector tvIONDRVStdInterruptHandler  = { (void *) IONDRVStdInterruptHandler,  0 };
static IOTVector tvIONDRVStdInterruptEnabler  = { (void *) IONDRVStdInterruptEnabler,  0 };
static IOTVector tvIONDRVStdInterruptDisabler = { (void *) IONDRVStdInterruptDisabler, 0 };
#endif

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

OSStatus
_eGetInterruptFunctions(
  InterruptSetID          set,
  InterruptMemberNumber   member,
  void **                 refCon,
  IOTVector **		  handler,
  IOTVector **            enabler,
  IOTVector **            disabler )
{
#if 0
    IONDRVInterruptSource *	source;
    OSStatus			err = noErr;

    assert( OSDynamicCast( IONDRVInterruptSet, set ));
    assert( member <= set->count );
    source = set->sources + member;

    if (refCon)
	*refCon   = source->refCon;
    if (handler)
	*handler  = source->handler;
    if (enabler)
	*enabler  = source->enabler;
    if (disabler)
	*disabler = source->disabler;

    return (err);
#endif
    return -1;
}

OSStatus
GetInterruptFunctions(
  InterruptSetID          set,
  InterruptMemberNumber   member,
  void **                 refCon,
  InterruptHandler *      handlerFunction,
  InterruptEnabler *      enableFunction,
  InterruptDisabler *     disableFunction)
{
#if 0
    IONDRVInterruptSource *	source;
    OSStatus			err = noErr;

    assert( OSDynamicCast( IONDRVInterruptSet, set ));
    assert( member <= set->count );
    source = set->sources + member;

    if (refCon)
	*refCon   = source->refCon;
    if (handlerFunction)
	*handlerFunction = (InterruptHandler)  source->handler->pc;
    if (enableFunction)
	*enableFunction  = (InterruptEnabler)  source->enabler->pc;
    if (disableFunction)
	*disableFunction = (InterruptDisabler) source->disabler->pc;

    return (err);
#endif
    return -1;
}

IOReturn
IONDRVInstallInterruptFunctions(void *		setID,
				UInt32 		member,
				void *		refCon,
				IOTVector *	handler,
				IOTVector *	enabler,
				IOTVector *	disabler )
{
#if 0
    IONDRVInterruptSet *	set;
    IONDRVInterruptSource *	source;
    OSStatus			err = noErr;

    set = (IONDRVInterruptSet *) setID;
    assert( OSDynamicCast( IONDRVInterruptSet, set ));
    if (member > set->count)
	return (paramErr);
    source = set->sources + member;

    source->refCon = refCon;
    if (handler)
	source->handler  = handler;
    if (enabler)
	source->enabler  = enabler;
    if (disabler)
	source->disabler = disabler;

    return (err);
#endif
    return -1;
}

OSStatus
_eInstallInterruptFunctions(
  InterruptSetID          setID,
  InterruptMemberNumber   member,
  void *                  refCon,
  IOTVector *             handler,
  IOTVector *             enabler,
  IOTVector *             disabler )
{
	//    return (IONDRVInstallInterruptFunctions(setID, member, refCon,
	//					    handler, enabler, disabler));

	return -1;
}

OSStatus
InstallInterruptFunctions(
  InterruptSetID          set,
  InterruptMemberNumber   member,
  void *                  refCon,
  InterruptHandler        handlerFunction,
  InterruptEnabler        enableFunction,
  InterruptDisabler       disableFunction)
{
#if 0
    IONDRVInterruptSource *	source;
    OSStatus			err = noErr;

    assert( OSDynamicCast( IONDRVInterruptSet, set ));
    if ((UInt32)member > set->count)
	return (paramErr);
    source = set->sources + member;

    source->refCon = refCon;
    if (handlerFunction)
    {
	source->handlerStore.pc = (void *) handlerFunction;
	source->handler = &source->handlerStore;
    }
    if (enableFunction)
    {
	source->enablerStore.pc = (void *) enableFunction;
	source->enabler = &source->enablerStore;
    }
    if (disableFunction)
    {
	source->disablerStore.pc = (void *) disableFunction;
	source->disabler = &source->disablerStore;
    }

    return (err);
#endif
    return -1;
}

OSStatus
EXP(CreateInterruptSet)(
  InterruptSetID          set,
  InterruptMemberNumber   parentMember,
  InterruptMemberNumber   setSize,
  InterruptSetID *        setID,
  InterruptSetOptions     options)
{
#if 0
    IONDRVInterruptSet *	newSet;
    IONDRVInterruptSource *	source;
    OSStatus			err = noErr;

    assert( OSDynamicCast( IONDRVInterruptSet, set ));
    assert( parentMember <= set->count );
    source = set->sources + parentMember;

    newSet = IONDRVInterruptSet::with( 0, options, setSize );
    assert( newSet );

    if (newSet)
    {
	for (UInt32 i = 1; i <= (UInt32) setSize; i++)
	{
	    source = newSet->sources + i;
	    source->handler 	= &tvIONDRVStdInterruptHandler;
	    source->enabler 	= &tvIONDRVStdInterruptEnabler;
	    source->disabler 	= &tvIONDRVStdInterruptDisabler;
	}
    }
    set->child = newSet;
    *setID = (InterruptSetID) newSet;

    return (err);
#endif
    return -1;
}

OSStatus
EXP(DeleteInterruptSet)( InterruptSetID set )
{
#if 0
    OSStatus			err = noErr;

    assert( OSDynamicCast( IONDRVInterruptSet, set ));

    set->release();

    return (err);
#endif
    return -1;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

extern FunctionEntry NameRegistryLibFuncs[];

extern FunctionEntry PCILibFuncs[];

static FunctionEntry VideoServicesLibFuncs[] =
    {
	MAKEFUNC( "VSLPrepareCursorForHardwareCursor",
		    EXP(VSLPrepareCursorForHardwareCursor)),
	MAKEFUNC( "VSLNewInterruptService", EXP(VSLNewInterruptService)),
	MAKEFUNC( "VSLDisposeInterruptService", EXP(VSLDisposeInterruptService)),
	MAKEFUNC( "VSLDoInterruptService", EXP(VSLDoInterruptService)),
	MAKEFUNC( "VSLSetDisplayConfiguration", EXP(VSLSetDisplayConfiguration)),
	MAKEFUNC( "VSLGestalt", EXP(VSLGestalt)),
	ENDFUNC()
    };

static FunctionEntry DriverServicesLibFuncs[] =
    {
	MAKEFUNC( "SynchronizeIO", EXP(SynchronizeIO)),
	MAKEFUNC( "SetProcessorCacheMode", EXP(SetProcessorCacheMode)),
	MAKEFUNC( "BlockCopy", EXP(BlockCopy)),
	MAKEFUNC( "BlockMove", EXP(BlockMove)),
	MAKEFUNC( "BlockMoveData", EXP(BlockMoveData)),

	MAKEFUNC( "BlockMoveDataUncached", EXP(BlockMoveDataUncached)),
	MAKEFUNC( "BlockMoveUncached", EXP(BlockMoveUncached)),

	MAKEFUNC( "BlockZero", EXP(BlockZero)),
	MAKEFUNC( "BlockZeroUncached", EXP(BlockZeroUncached)),

	MAKEFUNC( "CStrCopy", EXP(CStrCopy)),
	MAKEFUNC( "CStrCmp", EXP(CStrCmp)),
	MAKEFUNC( "CStrLen", EXP(CStrLen)),
	MAKEFUNC( "CStrCat", EXP(CStrCat)),
	MAKEFUNC( "CStrNCopy", EXP(CStrNCopy)),
	MAKEFUNC( "CStrNCmp", EXP(CStrNCmp)),
	MAKEFUNC( "CStrNCat", EXP(CStrNCat)),
	MAKEFUNC( "PStrCopy", EXP(PStrCopy)),
	MAKEFUNC( "PStrToCStr", EXP(PStrToCStr)),
	MAKEFUNC( "CStrToPStr", EXP(CStrToPStr)),

	MAKEFUNC( "PoolAllocateResident", EXP(PoolAllocateResident)),
	MAKEFUNC( "MemAllocatePhysicallyContiguous", EXP(PoolAllocateResident)),
	MAKEFUNC( "PoolDeallocate", EXP(PoolDeallocate)),

	MAKEFUNC( "UpTime", EXP(UpTime)),
	MAKEFUNC( "AbsoluteDeltaToDuration", EXP(AbsoluteDeltaToDuration)),
	MAKEFUNC( "AbsoluteToDuration", EXP(AbsoluteToDuration)),
	MAKEFUNC( "AddAbsoluteToAbsolute", EXP(AddAbsoluteToAbsolute)),
	MAKEFUNC( "SubAbsoluteFromAbsolute", EXP(SubAbsoluteFromAbsolute)),
	MAKEFUNC( "AddDurationToAbsolute", EXP(AddDurationToAbsolute)),
	MAKEFUNC( "NanosecondsToAbsolute", EXP(NanosecondsToAbsolute)),
	MAKEFUNC( "AbsoluteToNanoseconds", EXP(AbsoluteToNanoseconds)),
	MAKEFUNC( "DurationToAbsolute", EXP(DurationToAbsolute)),
	MAKEFUNC( "DelayForHardware", EXP(DelayForHardware)),
	MAKEFUNC( "DelayFor", EXP(DelayFor)),

	MAKEFUNC( "CurrentExecutionLevel", EXP(CurrentExecutionLevel)),
	MAKEFUNC( "IOCommandIsComplete", EXP(IOCommandIsComplete)),

	MAKEFUNC( "SysDebugStr", EXP(SysDebugStr)),
	MAKEFUNC( "SysDebug", EXP(SysDebug)),

	MAKEFUNC( "CompareAndSwap", EXP(CompareAndSwap)),

	MAKEFUNC( "CreateInterruptSet", EXP(CreateInterruptSet)),
	MAKEFUNC( "DeleteInterruptSet", EXP(DeleteInterruptSet)),
	MAKEFUNC( "GetInterruptFunctions", _eGetInterruptFunctions),
	MAKEFUNC( "InstallInterruptFunctions", _eInstallInterruptFunctions),
	ENDFUNC()
    };

static FunctionEntry ATIUtilsFuncs[] =
    {
	// Gossamer onboard ATI
	MAKEFUNC( "ATISetMBRES", EXP(ATISetMBRES)),
	MAKEFUNC( "ATISetMonitorTermination", EXP(ATISetMonitorTermination)),
	MAKEFUNC( "ATIIsAllInOne", EXP(ATIIsAllInOne)),
	ENDFUNC()
    };

// These are all out of spec

static FunctionEntry InterfaceLibFuncs[] =
    {
	// Apple control : XPRam and EgretDispatch
	MAKEFUNC( "CallUniversalProc", _eFail),
	MAKEFUNC( "CallOSTrapUniversalProc", _eCallOSTrapUniversalProc),
	MAKEFUNC( "BlockZero", EXP(BlockZero)),
	MAKEFUNC( "BlockZeroUncached", EXP(BlockZeroUncached)),

	// Apple chips65550
	//    MAKEFUNC( "NewRoutineDescriptor", _eCallOSTrapUniversalProc),
	//    MAKEFUNC( "DisposeRoutineDescriptor", _eNoErr),
	//    MAKEFUNC( "InsTime", _eInsTime),
	//    MAKEFUNC( "PrimeTime", _ePrimeTime),

	// Radius PrecisionColor 16
	MAKEFUNC( "CountADBs", _eNoErr),
	MAKEFUNC( "GetIndADB", _eGetIndADB),
	MAKEFUNC( "GetKeys", _eGetKeys),
	ENDFUNC()
    };

static FunctionEntry PrivateInterfaceLibFuncs[] =
    {
	// Apple chips65550
	    MAKEFUNC( "LMGetPowerMgrVars", _eLMGetPowerMgrVars ),
	    ENDFUNC()
    };

#define NUMLIBRARIES	7
const ItemCount IONumNDRVLibraries = NUMLIBRARIES;
LibraryEntry IONDRVLibraries[ NUMLIBRARIES ] =
    {
	{ "PCILib", PCILibFuncs },
	{ "VideoServicesLib", VideoServicesLibFuncs },
	{ "NameRegistryLib", NameRegistryLibFuncs },
	{ "DriverServicesLib", DriverServicesLibFuncs },

	// G3
	{ "ATIUtils", ATIUtilsFuncs },

	// out of spec stuff
	{ "InterfaceLib", InterfaceLibFuncs },
	{ "PrivateInterfaceLib", PrivateInterfaceLibFuncs }
    };


